# -*- coding: utf-8 -*-
# -*- mode: python -*-
from django.urls import re_path

from decide_host import views

urlpatterns = [
    re_path(r"^$", views.api_root, name="index"),
    re_path(r"^api/$", views.api_root, name="index"),
    re_path(r"^api/info/$", views.api_info, name="api-info"),
    re_path(r"^api/events/$", views.EventList.as_view(), name="event-list"),
    re_path(r"^api/trials/$", views.TrialList.as_view(), name="trial-list"),
    re_path(
        r"^api/controllers/$", views.ControllerList.as_view(), name="controller-list"
    ),
    re_path(
        r"^api/controllers/(?P<name>[\w-]+)/$",
        views.ControllerDetail.as_view(),
        name="controller-detail",
    ),
    re_path(
        r"^api/controllers/(?P<addr>[\w-]+)/events/$",
        views.ControllerEventList.as_view(),
        name="controller-event-list",
    ),
    re_path(r"^api/subjects/$", views.SubjectList.as_view(), name="subject-list"),
    re_path(
        r"^api/subjects/(?P<name>[\w-]+)/$",
        views.SubjectDetail.as_view(),
        name="subject-detail",
    ),
    re_path(
        r"^api/subjects/(?P<subject>[\w-]+)/trials/$",
        views.SubjectTrialList.as_view(),
        name="subject-trial-list",
    ),
    # fallthrough to 404
    re_path(r"^.*/$", views.notfound, name="notfound"),
]



# -*- coding: utf-8 -*-
# -*- mode: python -*-
from __future__ import unicode_literals

import datetime

from django.core.exceptions import ObjectDoesNotExist
from django.utils.encoding import smart_str
from django.utils.timezone import make_aware
from rest_framework import serializers
from decide_host.models import Controller, Component, Event, Subject, Trial

# ideally, I'd like to flatten the JSON so that all the data fields are at the
# top level. This simplifies the representation


class CreatableSlugRelatedField(serializers.SlugRelatedField):
    def to_internal_value(self, data):
        try:
            return self.get_queryset().get_or_create(**{self.slug_field: data})[0]
        except ObjectDoesNotExist:
            self.fail(
                "does_not_exist", slug_name=self.slug_field, value=smart_str(data)
            )
        except (TypeError, ValueError):
            self.fail("invalid")


class MicroDateTimeField(serializers.DateTimeField):
    """Class to handle floating point timestamps"""

    def to_internal_value(self, data):
        if isinstance(data, float):
            return make_aware(datetime.datetime.fromtimestamp(data))
        if isinstance(data, int):
            return make_aware(datetime.datetime.fromtimestamp(data * 1e-6))
        else:
            return super(MicroDateTimeField, self).to_internal_value(data)


class JSONFlattenMixin(object):
    """Flatens the specified related objects in this representation"""

    def to_representation(self, obj):
        """Move fields from data to top-level"""
        try:
            flatten_field = self.Meta.flatten
        except AttributeError:
            raise AssertionError(
                'Class {serializer_class} missing "Meta.flatten" attribute'.format(
                    serializer_class=self.__class__.__name__
                )
            )
        repr = super(JSONFlattenMixin, self).to_representation(obj)
        data = repr.pop(flatten_field)
        for key in data:
            repr[key] = data[key]
        return repr

    def to_internal_value(self, data):
        """Nest fields not in Meta.fields"""
        try:
            flatten_field = self.Meta.flatten
        except AttributeError:
            raise AssertionError(
                'Class {serializer_class} missing "Meta.flatten" attribute'.format(
                    serializer_class=self.__class__.__name__
                )
            )
        nested = {key: data[key] for key in data if key not in self.Meta.fields}
        not_nested = {key: data[key] for key in data if key in self.Meta.fields}
        internal = super().to_internal_value(not_nested)
        try:
            internal[flatten_field].update(nested)
        except KeyError:
            internal[flatten_field] = nested
        return internal


class ControllerSerializer(serializers.ModelSerializer):
    n_events = serializers.IntegerField(read_only=True)
    last_event_time = serializers.DateTimeField(read_only=True)

    class Meta:
        model = Controller
        fields = ("name", "last_event_time", "n_events")


class SubjectSerializer(serializers.ModelSerializer):
    n_trials = serializers.IntegerField(read_only=True)
    last_trial_time = serializers.DateTimeField(read_only=True)
    n_trials_today = serializers.IntegerField(read_only=True)

    class Meta:
        model = Subject
        fields = (
            "name",
            "last_trial_time",
            "n_trials_today",
            "n_trials",
        )


class EventSerializer(JSONFlattenMixin, serializers.ModelSerializer):
    addr = CreatableSlugRelatedField(
        queryset=Controller.objects.all(), slug_field="name"
    )
    name = CreatableSlugRelatedField(
        queryset=Component.objects.all(), slug_field="name"
    )
    time = MicroDateTimeField()
    data = serializers.JSONField(default=dict)

    class Meta:
        model = Event
        fields = ("addr", "name", "time", "data")
        flatten = "data"


class TrialSerializer(JSONFlattenMixin, serializers.ModelSerializer):
    addr = CreatableSlugRelatedField(
        queryset=Controller.objects.all(), slug_field="name"
    )
    name = CreatableSlugRelatedField(
        queryset=Component.objects.all(), slug_field="name"
    )
    subject = CreatableSlugRelatedField(
        queryset=Subject.objects.all(), slug_field="name"
    )
    time = MicroDateTimeField()
    data = serializers.JSONField(default=dict)

    class Meta:
        model = Trial
        fields = ("addr", "name", "subject", "time", "data")
        flatten = "data"



# -*- coding: utf-8 -*-
# -*- mode: python -*-
from __future__ import unicode_literals

from django.db.models import JSONField, Count, Max, Q
from django.utils import timezone
from django.db import models


class EventManager(models.Manager):
    def with_names(self):
        return self.select_related("addr", "name")


class Event(models.Model):
    """An event is any state change in a connected controller"""

    id = models.AutoField(primary_key=True)
    addr = models.ForeignKey(
        "Controller",
        on_delete=models.PROTECT,
        help_text="the controller that generated the event",
    )
    name = models.ForeignKey(
        "Component", on_delete=models.PROTECT, help_text="the name of the component"
    )
    time = models.DateTimeField(db_index=True)
    data = JSONField()
    objects = EventManager()

    def __str__(self):
        return "%s:%s @ %s" % (self.addr, self.name, self.time.isoformat())

    class Meta:
        unique_together = ("addr", "name", "time")
        indexes = [models.Index(fields=["addr", "-time"], name="addr_time_desc_idx")]
        ordering = ("-time",)


class TrialManager(models.Manager):
    def with_names(self):
        return self.select_related("addr", "subject", "name")


class Trial(models.Model):
    """A trial is one set of events in an experimental paradigm"""

    id = models.AutoField(primary_key=True)
    addr = models.ForeignKey(
        "Controller", on_delete=models.PROTECT, help_text="the controller for the trial"
    )
    name = models.ForeignKey(
        "Component",
        on_delete=models.PROTECT,
        help_text="the experiment paradigm (e.g., shape)",
    )
    subject = models.ForeignKey(
        "Subject", on_delete=models.PROTECT, help_text="the experimental subject"
    )
    time = models.DateTimeField(db_index=True)
    data = JSONField()
    objects = TrialManager()

    def __str__(self):
        return "%s:%s @ %s" % (self.addr, self.subject, self.time.isoformat())

    class Meta:
        unique_together = ("name", "subject", "time")
        indexes = [
            models.Index(fields=["subject", "-time"], name="subject_time_desc_idx")
        ]
        ordering = ("-time",)


class ControllerManager(models.Manager):
    def with_counts(self):
        return self.annotate(
            n_events=Count("event"), last_event_time=Max("event__time")
        )


class Controller(models.Model):
    """Represents a controller connected to this host"""

    id = models.AutoField(primary_key=True)
    name = models.SlugField(max_length=32, unique=True)
    objects = ControllerManager()

    def __str__(self):
        return self.name

    class Meta:
        ordering = ("name",)


class Component(models.Model):
    """Represents a component in a controller"""

    # the main reason for having this as a separate model rather than a string
    # field is to reduce data size and enhance indexing
    id = models.AutoField(primary_key=True)
    name = models.SlugField(max_length=32, unique=True)

    def __str__(self):
        return self.name

    class Meta:
        ordering = ("name",)


class SubjectManager(models.Manager):
    def with_counts(self):
        today_start = timezone.localtime(timezone.now()).replace(
            hour=0, minute=0, second=0
        )
        return self.annotate(
            n_trials=Count("trial"),
            last_trial_time=Max("trial__time"),
            n_trials_today=Count("trial", filter=Q(trial__time__gte=today_start)),
        )


class Subject(models.Model):
    """Represents an experimental subject"""

    id = models.AutoField(primary_key=True)
    name = models.SlugField(max_length=36, unique=True)
    objects = SubjectManager()

    def n_trials_today(self):
        today_start = timezone.localtime(timezone.now()).replace(
            hour=0, minute=0, second=0
        )
        today_end = today_start.replace(hour=23, minute=59, second=59)
        return self.trial_set.filter(time__gte=today_start, time__lte=today_end).count()

    def __str__(self):
        return self.name

    class Meta:
        ordering = ("name",)



# -*- coding: utf-8 -*-
# -*- mode: python -*-
from __future__ import unicode_literals

from django.contrib import admin
from decide_host.models import Event, Trial, Controller, Component, Subject

for model in (Event, Trial, Controller, Component, Subject):
    admin.site.register(model)



# -*- coding: utf-8 -*-
# -*- mode: python -*-

import ipaddress
import logging

from django.conf import settings
from django.shortcuts import get_object_or_404
from rest_framework import generics, status, permissions
from rest_framework.response import Response
from rest_framework.decorators import api_view
from rest_framework.reverse import reverse
from drf_link_header_pagination import LinkHeaderCursorPagination
from django_filters import rest_framework as filters

from decide_host import __version__, api_version
from decide_host import models
from decide_host import serializers

logger = logging.getLogger(__name__)


class LinkHeaderCursorPaginationByTimestamp(LinkHeaderCursorPagination):
    ordering = "-time"


@api_view(["HEAD", "GET"])
def api_root(request, format=None):
    urls = {
        "info": reverse("api-info", request=request, format=format),
        "events": reverse("event-list", request=request, format=format),
        "trials": reverse("trial-list", request=request, format=format),
        "controllers": reverse("controller-list", request=request, format=format),
        "subjects": reverse("subject-list", request=request, format=format),
    }
    headers = {
        "Link": ", ".join(f'<{uri}>; rel="{name}"' for name, uri in urls.items())
    }
    return Response(urls, headers=headers)


@api_view(["GET"])
def api_info(request, format=None):
    return Response(
        {
            "host": "django-decide-host",
            "host_version": __version__,
            "api_version": api_version,
        }
    )


@api_view(["GET"])
def notfound(request, format=None):
    return Response({"detail": "not found"}, status=status.HTTP_404_NOT_FOUND)


class IsAuthorizedSubnetOrReadOnly(permissions.BasePermission):
    message = "read-only access"

    def has_permission(self, request, view):
        if request.method in permissions.SAFE_METHODS:
            return True
        else:
            ip_addr = request.META["REMOTE_ADDR"]
            for subnet in settings.DECIDE_HOST.get("READWRITE_SUBNETS", []):
                if ipaddress.ip_address(ip_addr) in ipaddress.ip_network(subnet):
                    return True
            return False


class DataFieldFilterMixin(object):
    """Provides filtering based on components of the data JSONField"""

    def filter_queryset(self, queryset):
        qs = super().filter_queryset(queryset)
        # this could be a little dangerous b/c we're letting the user design queries
        mq = {k: v for k, v in self.request.GET.items() if k.startswith("data__")}
        return qs.filter(**mq)


class EventFilter(filters.FilterSet):
    addr = filters.CharFilter(
        field_name="addr__name", label="addr", lookup_expr="iexact"
    )
    name = filters.CharFilter(
        field_name="name__name", label="name", lookup_expr="iexact"
    )
    date = filters.DateFromToRangeFilter(field_name="time")

    class Meta:
        model = models.Event
        fields = {"time": ["exact", "date"]}


class TrialFilter(filters.FilterSet):
    addr = filters.CharFilter(
        field_name="addr__name", label="addr", lookup_expr="iexact"
    )
    name = filters.CharFilter(
        field_name="name__name", label="name", lookup_expr="iexact"
    )
    subject = filters.CharFilter(
        field_name="subject__name", label="subject", lookup_expr="iexact"
    )
    nocomment = filters.BooleanFilter(
        field_name="data__comment", label="nocomment", lookup_expr="isnull"
    )
    date = filters.DateFromToRangeFilter(field_name="time")

    class Meta:
        model = models.Trial
        fields = {"time": ["exact", "date"]}


class EventList(DataFieldFilterMixin, generics.ListCreateAPIView):
    queryset = models.Event.objects.with_names()
    serializer_class = serializers.EventSerializer
    filter_backends = (filters.DjangoFilterBackend,)
    filterset_class = EventFilter
    pagination_class = LinkHeaderCursorPaginationByTimestamp
    permission_classes = (IsAuthorizedSubnetOrReadOnly,)


class TrialList(DataFieldFilterMixin, generics.ListCreateAPIView):
    """Records of trials and trial-related comments.

    This endpoint returns a list of records from the database of trial records.
    Use query parameters to restrict the number of items to a reasonable subset.

    Basic filters: `name` (of the procedure), `addr` (of the controller), and
    `subject`.

    Date-based filters: `time__date` for a specific date, `date_before` and
    `date_after` to specify a range. Ranges are exclusive. Format dates as `YYYY-MM-DD`.

    Exclude comments with the query `nocomment=true`.

    You can filter on other fields of the record, but these need to be prefaced
    by `data__` due to the way they're stored in the database. For example, to
    restrict returned records to ones in which `correct` was `True`, add the
    query `data__correct=True`.

    Multiple queries produce a more restrictive filter.

    Example: "?subject=P24&date_after=2022-02-01&date_before=2022-02-28"

    """

    queryset = models.Trial.objects.with_names()
    serializer_class = serializers.TrialSerializer
    filter_backends = (filters.DjangoFilterBackend,)
    filterset_class = TrialFilter
    pagination_class = LinkHeaderCursorPaginationByTimestamp
    permission_classes = (IsAuthorizedSubnetOrReadOnly,)


class ControllerList(generics.ListAPIView):
    queryset = models.Controller.objects.with_counts().order_by("-last_event_time")
    serializer_class = serializers.ControllerSerializer


class ControllerDetail(generics.RetrieveAPIView):
    lookup_field = "name"
    queryset = models.Controller.objects.with_counts()
    serializer_class = serializers.ControllerSerializer

    def retrieve(self, request, **kwargs):
        response = super().retrieve(request, **kwargs)
        # add link to trials in header
        uri = reverse("controller-event-list", args=[kwargs["name"]], request=request)
        response["Link"] = f'<{uri}>; rel="events"'
        return response


class ControllerEventList(DataFieldFilterMixin, generics.ListAPIView):
    serializer_class = serializers.EventSerializer
    filter_backends = (filters.DjangoFilterBackend,)
    pagination_class = LinkHeaderCursorPaginationByTimestamp
    filterset_class = EventFilter

    def get_object(self):
        return get_object_or_404(models.Controller, name=self.kwargs["addr"])

    def get_queryset(self):
        addr = self.get_object()
        return addr.event_set.with_names()


class SubjectList(generics.ListAPIView):
    queryset = models.Subject.objects.with_counts().order_by("-last_trial_time")
    serializer_class = serializers.SubjectSerializer


class SubjectDetail(generics.RetrieveAPIView):
    lookup_field = "name"
    queryset = models.Subject.objects.with_counts()
    serializer_class = serializers.SubjectSerializer

    def retrieve(self, request, **kwargs):
        response = super().retrieve(request, **kwargs)
        # add link to trials in header
        uri = reverse("subject-trial-list", args=[kwargs["name"]], request=request)
        response["Link"] = f'<{uri}>; rel="trials"'
        return response


class SubjectTrialList(DataFieldFilterMixin, generics.ListAPIView):
    serializer_class = serializers.TrialSerializer
    filter_backends = (filters.DjangoFilterBackend,)
    pagination_class = LinkHeaderCursorPaginationByTimestamp
    filterset_class = TrialFilter

    def get_object(self):
        return get_object_or_404(models.Subject, name=self.kwargs["subject"])

    def get_queryset(self):
        subj = self.get_object()
        return subj.trial_set.with_names()



# -*- coding: utf-8 -*-
# -*- mode: python -*-
""" Import line-delimited records into trial objects using the Django ORM """

from django.core.management.base import BaseCommand, CommandError
from decide_host.models import Trial
from decide_host.serializers import TrialSerializer

import json


class Command(BaseCommand):
    help = "import trials from line-delimited JSON (jsonl) files"

    def add_arguments(self, parser):
        parser.add_argument("-n", "--name", help="name of procedure (required)", required=True)
        parser.add_argument("-a", "--addr", help="controller name (required)", required=True)
        parser.add_argument('files', nargs="+", help="files to import")

    def handle(self, *args, **options):
        for path in options['files']:
            with open(path, "rU") as fp:
                n = 0
                for i, line in enumerate(fp):
                    record = json.loads(line.strip())
                    record['addr'] = options['addr']
                    record['name'] = options['name']
                    serializer = TrialSerializer(data=record)
                    if serializer.is_valid():
                        n += 1
                        serializer.save()
                    else:
                        self.stdout.write("%s, record %d invalid: %s" % (path, i, serializer.errors))
                self.stdout.write("%s: imported %d record(s)" % (path, n))
